import json
from enum import StrEnum
from typing import Any, Optional

from pydantic import BaseModel, Field, model_validator

#
# Enum definitions
#


class OpenAPISchemaType(StrEnum):
    OBJECT = "object"
    STRING = "string"
    INTEGER = "integer"
    FLOAT = "float"
    BOOLEAN = "boolean"
    ARRAY = "array"
    # Typically used to indicate an optional type within an anyOf statement
    NULL = "null"


class ParameterLocationType(StrEnum):
    # https://swagger.io/specification: Parameter Object
    PATH = "path"
    QUERY = "query"

    # https://swagger.io/docs/specification/authentication/cookie-authentication/
    COOKIE = "cookie"
    HEADER = "header"


class ActionType(StrEnum):
    GET = "get"
    POST = "post"
    PUT = "put"
    PATCH = "patch"
    DELETE = "delete"


#
# Nested schemas inside OpenAPI definitions
#


class OpenAPIProperty(BaseModel):
    """
    A property is the core wrapper for OpenAPI model objects. It allows users to recursively
    define data structures based on their type. Each property can have a list of subproperties, alongside
    the constraints of how they're used (ie. required attributes, union of attributes, etc).

    """

    title: str | None = None
    description: str | None = None
    properties: dict[str, "OpenAPIProperty"] = {}
    additionalProperties: Optional["OpenAPIProperty"] = None
    required: list[str] = []

    # Just specified on the leaf object
    format: str | None = None

    # Self-contained type: object, int, etc
    variable_type: OpenAPISchemaType | None = Field(alias="type", default=None)
    # Reference to another type
    ref: str | None = Field(alias="$ref", default=None)
    # Array of another type
    items: Optional["OpenAPIProperty"] = None
    # Enum type
    enum: list[Any] | None = None

    default: Any | None = None

    # Pointer to multiple possible subtypes
    anyOf: list["OpenAPIProperty"] = []

    model_config = {"populate_by_name": True}

    # Validator to ensure that one of the optional values is set
    @model_validator(mode="after")
    def check_provided_value(self: "OpenAPIProperty") -> "OpenAPIProperty":
        if not any([self.variable_type, self.ref, self.items, self.anyOf, self.enum]):
            raise ValueError(
                "One of variable_type, $ref, anyOf, enum, or items must be set"
            )
        return self

    @classmethod
    def from_meta(
        cls,
        title: str | None = None,
        description: str | None = None,
        properties: dict[str, "OpenAPIProperty"] = {},
        additional_properties: Optional["OpenAPIProperty"] = None,
        required: list[str] = [],
        format: str | None = None,
        variable_type: OpenAPISchemaType | None = None,
        ref: str | None = None,
        items: Optional["OpenAPIProperty"] = None,
        enum: list[Any] | None = None,
        anyOf: list["OpenAPIProperty"] = [],
    ) -> "OpenAPIProperty":
        return cls.model_validate(
            {
                "title": title,
                "description": description,
                "properties": properties,
                "additionalProperties": additional_properties,
                "required": required,
                "format": format,
                "type": variable_type,
                "$ref": ref,
                "items": items,
                "enum": enum,
                "anyOf": anyOf,
            }
        )

    def __hash__(self):
        # Normally we would make use of a frozen BaseClass to enable hashing, but since
        # dictionaries are included in the payload here an easier way is just to convert
        # to a JSON string and hash that.
        # We make sure to order the strings since otherwise the hash risks being different
        # despite having the same values
        def sort_json(obj):
            if isinstance(obj, dict):
                return sorted((k, sort_json(v)) for k, v in obj.items())
            else:
                return obj

        return hash(json.dumps(sort_json(self.model_dump())))


class ContentDefinition(BaseModel):
    class Reference(BaseModel):
        ref: str | None = Field(default=None, alias="$ref")

        model_config = {"populate_by_name": True}

        @classmethod
        def from_meta(cls, ref: str) -> "ContentDefinition.Reference":
            # Workaround to pyright overriding the instance method with
            # the alias, but we can't create attributes with these dynamic types
            return cls.model_validate({"$ref": ref})

    schema_ref: Reference = Field(alias="schema")

    model_config = {"populate_by_name": True}

    @classmethod
    def from_meta(cls, schema_ref: Reference) -> "ContentDefinition":
        return cls.model_validate({"schema": schema_ref})


class ContentBodyDefinition(BaseModel):
    # original key is a `content: { content_type: {schema: SchemaDefinition }}`
    content_type: str
    content_schema: ContentDefinition

    # Requests will typically provider their required status, requests will not.
    # Default these to True since they are required within the scope of the request
    # that is provided.
    required: bool = True

    @model_validator(mode="before")
    def explode_content_dictionary(cls, data: Any) -> Any:
        # If we're being invoked programatically, we will have the required fields
        programatic_construction = data.get("content_type") and data.get(
            "content_schema"
        )

        if programatic_construction:
            return data

        # If we're being invoked from a JSON payload, we expect a content dictionary with a single
        # key/value that provides the specification for content type/content. Explode it so it maps
        # to our variables.
        if "content" not in data or not isinstance(data["content"], dict):
            raise ValueError("ContentBodyDefinition.content must be a dict")

        # We only support a single content type for now
        if len(data["content"]) != 1:
            raise ValueError(
                "RequestBodyDefinition.content must have a single content type"
            )

        data["content_type"], data["content_schema"] = list(data["content"].items())[0]
        return data


class URLParameterDefinition(BaseModel):
    name: str
    in_location: ParameterLocationType = Field(alias="in")
    schema_ref: OpenAPIProperty = Field(alias="schema")
    required: bool

    model_config = {"populate_by_name": True}

    @classmethod
    def from_meta(
        cls,
        name: str,
        in_location: ParameterLocationType,
        schema_ref: OpenAPIProperty,
        required: bool,
    ) -> "URLParameterDefinition":
        return cls.model_validate(
            {
                "name": name,
                "in": in_location,
                "schema": schema_ref,
                "required": required,
            }
        )


class ActionDefinition(BaseModel):
    action_type: ActionType

    summary: str
    operationId: str
    # Parameters injected into the URL path
    parameters: list[URLParameterDefinition] = []

    # { status_code: ResponseDefinition }
    responses: dict[str, ContentBodyDefinition]
    requestBody: ContentBodyDefinition | None = None

    # Custom Mountaineer event types specified in the OpenAPI schema
    # These should all have defaults since they're optional
    media_type: str | None = None
    is_raw_response: bool = False


class EndpointDefinition(BaseModel):
    actions: list[ActionDefinition] = []

    @model_validator(mode="before")
    def inject_action_type(cls, data: Any) -> dict[str, Any]:
        """
        OpenAPI often defines metadata in a dict structure where the key is relevant
        to the behavior of the values. In our pipeline we want to be able to pass
        around ActionDefinitions to fully generate a resulting type action. We migrate
        the key-metadata into the actual definition itself.

        """
        if not isinstance(data, dict):
            raise ValueError("EndpointDefinition must be a dict")

        for action_type, payload in data.items():
            payload["action_type"] = action_type

        return {"actions": data.values()}


#
# Top-level OpenAPI definition entrypoints. These parse the high level spec.
#


class OpenAPISchema(OpenAPIProperty):
    """
    Defines the expected format for model-only definition schemas. This
    is the output when Pydantic is called with `model_json_schema`.

    """

    defs: dict[str, OpenAPIProperty] = Field(alias="$defs", default_factory=dict)


class OpenAPIDefinition(BaseModel):
    """
    Defines the spec for a whole OpenAPI API definition. This mirrors what FastAPI
    outputs as the /openapi.json endpoint.

    """

    class Components(BaseModel):
        schemas: dict[str, OpenAPIProperty]

    # { path: { action: ActionDefinition }}
    paths: dict[str, EndpointDefinition]
    components: Components = Components(schemas={})


#
# Helper methods
#


def get_types_from_parameters(schema: OpenAPIProperty):
    """
    Handle potentially complex types from the parameter schema, like the case
    of optional fields.

    """
    # Recursively gather all of the types that might be nested
    if schema.variable_type:
        yield schema.variable_type

    for property in schema.properties.values():
        yield from get_types_from_parameters(property)

    if schema.additionalProperties:
        yield from get_types_from_parameters(schema.additionalProperties)

    if schema.items:
        yield from get_types_from_parameters(schema.items)

    if schema.anyOf:
        for one_of in schema.anyOf:
            yield from get_types_from_parameters(one_of)

    # We don't expect $ref values in the URL schema, if we do then the parsing
    # is likely incorrect
    if schema.ref:
        raise ValueError(f"Unexpected $ref in URL schema: {schema.ref}")



from typing import Any

from inflection import underscore

from mountaineer.client_builder.openapi import (
    ActionDefinition,
    ContentDefinition,
    OpenAPIDefinition,
    ParameterLocationType,
)
from mountaineer.client_builder.typescript import (
    TSLiteral,
    get_typehint_for_parameter,
    python_payload_to_typescript,
)
from mountaineer.constants import STREAM_EVENT_TYPE


class OpenAPIToTypescriptActionConverter:
    """
    Parse utilities and typescript construction for building actions
    based on the defined endpoint OpenAPI specs.

    """

    def convert(self, openapi: dict[str, Any]) -> tuple[dict[str, str], list[str]]:
        """
        Our conversion pipeline focuses on creating the action definitions of one file.

        :return {function_name: function_body}, imports required by the function bodies

        """
        schema = OpenAPIDefinition(**openapi)
        output_actions: dict[str, str] = {}
        output_errors: dict[str, str] = {}
        all_required_types: set[str] = set()
        for url, endpoint_definition in schema.paths.items():
            for action, method_name in zip(
                endpoint_definition.actions,
                self.get_method_names(url, endpoint_definition.actions),
            ):
                rendered_str, required_types = self.build_action(
                    url, action, method_name
                )
                error_strs, error_types = self.build_error(action)

                output_actions[method_name] = rendered_str
                output_errors.update(error_strs)
                all_required_types.update(required_types)
                all_required_types.update(error_types)

        return {**output_actions, **output_errors}, list(all_required_types)

    def build_action(self, url: str, action: ActionDefinition, method_name: str):
        """
        Builds the action function. This should be more-or-less compatible with the common
        server fetch provided by `openapi-typescript-codegen`. Example return value:

        public static my_action({
            param_id,
            requestBody,
        }: {
            param_id: string;
            requestBody: MyActionRequest;
        }): CancelablePromise<any> {
            return __request(OpenAPI, {
                method: 'POST',
                url: '/{param_id}/my_action',
                path: {
                    param_id,
                },
                body: requestBody,
                mediaType: 'application/json',
                errors: {
                    422: ValidationErrorException,
                },
            });
        }
        """
        arguments, response_types = self.build_action_payload(url, action)
        parameters, request_types = self.build_action_parameters(action)

        lines: list[str] = []

        response_type_template: str
        if action.media_type == STREAM_EVENT_TYPE:
            response_type_template = (
                "Promise<AsyncGenerator<{response_type}, void, unknown>>"
            )
        else:
            response_type_template = "Promise<{response_type}>"

        if action.is_raw_response:
            response_type = response_type_template.format(response_type="Response")
        else:
            response_type = response_type_template.format(
                response_type=" | ".join(response_types)
            )

        lines.append(
            f"export const {method_name} = ({parameters}): {response_type} => {{\n"
            + "return __request(\n"
            + arguments
            + "\n);\n"
            + "}"
        )
        return "\n".join(lines), list(set(request_types + response_types))

    def build_error(self, action: ActionDefinition):
        """
        Build an error class that wraps the typehinted error contents payload.
        """
        error_classes: dict[str, str] = {}
        required_types: list[str] = []
        for error_code, response in action.responses.items():
            status_int = int(error_code)
            if self.status_code_is_valid(status_int):
                continue
            if not response.content_schema.schema_ref.ref:
                continue

            model_name = response.content_schema.schema_ref.ref.split("/")[-1]
            error_classes[
                model_name
            ] = f"export class {self.get_exception_class_name(model_name)} extends FetchErrorBase<{model_name}> {{}}"
            required_types.append(model_name)

        return error_classes, required_types

    def build_action_parameters(self, action: ActionDefinition):
        parameters_dict: dict[Any, Any] = {}
        typehint_dict: dict[Any, Any] = {}
        request_types: list[str] = []

        for parameter in action.parameters:
            typehint_key, typehint_value = get_typehint_for_parameter(parameter)
            if parameter.in_location in {
                ParameterLocationType.COOKIE,
                ParameterLocationType.HEADER,
            }:
                continue
            parameters_dict[parameter.name] = TSLiteral(parameter.name)
            typehint_dict[typehint_key] = typehint_value

        if (
            action.requestBody is not None
            and action.requestBody.content_schema.schema_ref.ref
        ):
            # We expect that the interface defined within the /models.ts will have the same name as
            # the model in the current OpenAPI spec
            model_name = action.requestBody.content_schema.schema_ref.ref.split("/")[-1]
            parameters_dict["requestBody"] = TSLiteral("requestBody")
            typehint_dict[TSLiteral("requestBody")] = TSLiteral(model_name)
            request_types.append(model_name)

        if not parameters_dict:
            # Empty query parameter
            return "", request_types

        parameters_str = python_payload_to_typescript(parameters_dict)
        typehint_str = python_payload_to_typescript(typehint_dict)

        return f"{parameters_str}: {typehint_str}", request_types

    def build_action_payload(self, url: str, action: ActionDefinition):
        # Since our typescript common functions have variable inputs here, it's cleaner
        # to put them into a dictionary and format whatever made it in as a flat
        # input list.
        response_types: list[str] = []
        common_params: dict[str, Any] = {
            "method": action.action_type.upper(),
            "url": url,
            "path": {},
            "query": {},
            "errors": {},
        }

        if action.requestBody is not None:
            common_params["body"] = TSLiteral("requestBody")
            common_params["mediaType"] = action.requestBody.content_type

        if action.parameters:
            for parameter in action.parameters:
                if parameter.in_location == ParameterLocationType.PATH:
                    common_params["path"][parameter.name] = TSLiteral(parameter.name)
                elif parameter.in_location == ParameterLocationType.QUERY:
                    common_params["query"][parameter.name] = TSLiteral(parameter.name)
                elif parameter.in_location in {
                    ParameterLocationType.COOKIE,
                    ParameterLocationType.HEADER,
                }:
                    # No-op, cookies will be sent automatically by fetch()
                    continue
                else:
                    raise NotImplementedError(
                        f"Parameter location {parameter.in_location} not supported"
                    )

        for status_code, response_definition in action.responses.items():
            status_int = int(status_code)
            if self.status_code_is_valid(status_int):
                # OK response, we can specify the expected response type
                # Multiple OK responses for a single action is unusual, but again we support it
                if not action.is_raw_response:
                    response_types.append(
                        self.get_typescript_name_from_content_definition(
                            response_definition.content_schema
                        )
                    )
            else:
                error_typehint = self.get_typescript_name_from_content_definition(
                    response_definition.content_schema
                )
                common_params["errors"][status_int] = TSLiteral(
                    # Provide a mapping to the error class
                    self.get_exception_class_name(error_typehint)
                )

        # Remove the optional keys that don't have any values
        for optional_parameter in ["errors", "path", "query"]:
            if not common_params[optional_parameter]:
                del common_params[optional_parameter]

        # Support for server-events
        if action.media_type == STREAM_EVENT_TYPE:
            common_params["eventStreamResponse"] = True

        # Support for raw responses
        if action.is_raw_response:
            common_params["outputFormat"] = "raw"

        return python_payload_to_typescript(common_params), response_types

    def get_method_names(self, url: str, actions: list[ActionDefinition]):
        # By convention, the last part of the URL is the method name
        base_method_name = underscore(url.split("/")[-1])

        method_names: list[str] = []

        # For the most part, actions will only have one type of request parameter associated
        # with them but this might not always be true. In the cases of a conflict we can
        # uniquely define the typescript name by appending the action type
        if len(actions) > 1:
            # By the time we get to this function, none of the passed actions should
            # have null action types
            for action in actions:
                if action.action_type is None:
                    raise ValueError(f"Action {action} is missing an action type")
                method_names.append(f"{base_method_name}_{action.action_type.lower()}")
        else:
            method_names.append(base_method_name)

        return method_names

    def get_typescript_name_from_content_definition(
        self, definition: ContentDefinition
    ):
        if not definition.schema_ref.ref:
            raise ValueError(
                f"Content definition {definition} does not have a schema reference"
            )
        return definition.schema_ref.ref.split("/")[-1]

    def status_code_is_valid(self, status_code: int):
        return status_code >= 200 and status_code < 300

    def get_exception_class_name(self, exception_typehint: str):
        """
        Given an error like "HTTPValidationError", responds with a class name of
        "HTTPValidationErrorException"

        """
        return f"{exception_typehint}Exception"



"""
Generator for TypeScript interfaces from OpenAPI specifications.
"""
from typing import Any, Dict, Iterator, Type, get_args, get_origin

from inflection import camelize
from pydantic import BaseModel, create_model

from mountaineer.annotation_helpers import get_value_by_alias, yield_all_subtypes
from mountaineer.client_builder.openapi import (
    OpenAPIProperty,
    OpenAPISchema,
    OpenAPISchemaType,
)
from mountaineer.client_builder.typescript import (
    TSLiteral,
    map_openapi_type_to_ts,
    python_payload_to_typescript,
)


class OpenAPIToTypescriptSchemaConverter:
    """
    Transform a pydantic.BaseModel into a TypeScript interface, by using
    OpenAPI as an intermediate layer. This also allows client callers to support
    generating interfaces from other OpenAPI-compliant schemas.

    """

    def __init__(self, export_interface: bool = False):
        self.export_interface = export_interface

    def get_model_json_schema(self, model: Type[BaseModel]):
        """
        By default pydantic will still include exclude=True parameters in the
        OpenAPI schema. This helper function creates a synthetic model
        before conversion so we exclude these unnecessary parameters.

        """
        self.validate_typescript_candidate(model)

        include_fields = {
            field: (field_info.annotation, field_info)
            for field, field_info in model.model_fields.items()
            if not field_info.exclude
        }

        synthetic_model = create_model(
            model.__name__,
            __config__=model.model_config,
            **include_fields,  # type: ignore
        )

        return synthetic_model.model_json_schema()

    def convert_schema_to_typescript(
        self,
        parsed_spec: OpenAPISchema,
        defaults_are_required: bool = False,
        all_fields_required: bool = False,
    ):
        # Fetch all the dependent models
        all_models = list(self.gather_all_models(parsed_spec))

        return {
            model.title: self.convert_schema_to_interface(
                model,
                base=parsed_spec,
                defaults_are_required=defaults_are_required,
                all_fields_required=all_fields_required,
            )
            for model in all_models
            if model.title and model.title.strip()
        }

    def gather_all_models(self, base: OpenAPISchema):
        """
        Return all unique models that are used in the given OpenAPI schema. This allows clients
        to build up all of the dependencies that the core model needs.

        :param base: The core OpenAPI Schema
        """

        def walk_models(property: OpenAPIProperty) -> Iterator[OpenAPIProperty]:
            if (
                property.variable_type == OpenAPISchemaType.OBJECT
                or property.enum is not None
            ):
                yield property
            if property.ref is not None:
                yield from walk_models(self.resolve_ref(property.ref, base))
            if property.items:
                yield from walk_models(property.items)
            if property.anyOf:
                for prop in property.anyOf:
                    yield from walk_models(prop)
            for prop in property.properties.values():
                yield from walk_models(prop)
            if property.additionalProperties:
                yield from walk_models(property.additionalProperties)

        return list(set(walk_models(base)))

    def resolve_ref(self, ref: str, base: BaseModel) -> OpenAPIProperty:
        """
        Resolve a $ref that points to a propery-compliant schema in the same document. If this
        ref points somewhere else in the document (that is valid but not a data model) than we
        raise a ValueError.

        """
        current_obj = base
        for part in ref.split("/"):
            if part == "#":
                current_obj = base
            else:
                try:
                    current_obj = get_value_by_alias(current_obj, part)
                except AttributeError as e:
                    raise AttributeError(
                        f"Invalid $ref, couldn't resolve path: {ref}"
                    ) from e
        if not isinstance(current_obj, OpenAPIProperty):
            raise ValueError(f"Resolved $ref is not a valid OpenAPIProperty: {ref}")
        return current_obj

    def convert_schema_to_interface(
        self,
        model: OpenAPIProperty,
        base: BaseModel,
        defaults_are_required: bool,
        all_fields_required: bool,
    ):
        if model.variable_type == OpenAPISchemaType.OBJECT:
            return self._convert_object_to_interface(
                model,
                base,
                defaults_are_required=defaults_are_required,
                all_fields_required=all_fields_required,
            )
        elif model.enum is not None:
            return self._convert_enum_to_interface(model)
        else:
            raise ValueError(f"Unknown model type: {model}")

    def _convert_object_to_interface(
        self,
        model: OpenAPIProperty,
        base: BaseModel,
        defaults_are_required: bool,
        all_fields_required: bool,
    ):
        fields = []

        # We have to support arrays with one and multiple values
        def walk_array_types(prop: OpenAPIProperty) -> Iterator[str]:
            if prop.ref:
                yield self.get_typescript_interface_name(
                    self.resolve_ref(prop.ref, base=base)
                )
            elif prop.items:
                yield from walk_array_types(prop.items)
            elif prop.anyOf:
                for sub_prop in prop.anyOf:
                    yield from walk_array_types(sub_prop)
            elif prop.additionalProperties:
                # OpenAPI doesn't specify the type of the keys since JSON forces them to be strings
                # By the time we get to this function we should have called validate_typescript_candidate
                sub_types = " | ".join(walk_array_types(prop.additionalProperties))
                yield f"Record<{map_openapi_type_to_ts(OpenAPISchemaType.STRING)}, {sub_types}>"
            elif prop.variable_type:
                yield map_openapi_type_to_ts(prop.variable_type)

        for prop_name, prop_details in model.properties.items():
            is_required = (
                (prop_name in model.required)
                or (defaults_are_required and prop_details.default is not None)
                or all_fields_required
            )

            ts_type = (
                map_openapi_type_to_ts(prop_details.variable_type)
                if prop_details.variable_type
                else None
            )

            annotation_str = " | ".join(set(walk_array_types(prop_details)))
            ts_type = (
                ts_type.format(types=annotation_str) if ts_type else annotation_str
            )

            if prop_details.description:
                fields.append("  /**")
                fields.append(f"   * {prop_details.description}")
                fields.append("   */")

            fields.append(f"  {prop_name}{'?' if not is_required else ''}: {ts_type};")

        interface_body = "\n".join(fields)
        interface_full = f"interface {self.get_typescript_interface_name(model)} {{\n{interface_body}\n}}"

        if self.export_interface:
            interface_full = f"export {interface_full}"

        return interface_full

    def _convert_enum_to_interface(self, model: OpenAPIProperty):
        fields: dict[str, Any] = {}

        if not model.enum:
            raise ValueError(f"Model {model} is not an enum")

        for enum_value in model.enum:
            # If the enum is an integer, we need to escape it
            enum_key: str
            if isinstance(enum_value, (int, float)):
                enum_key = f"Value__{enum_value}"
            elif isinstance(enum_value, str):
                enum_key = camelize(enum_value, uppercase_first_letter=True)
            else:
                raise ValueError(f"Invalid enum value: {enum_value}")

            fields[TSLiteral(enum_key)] = enum_value

        # Enums use an equal assignment syntax
        interface_body = python_payload_to_typescript(fields).replace(":", " =")
        interface_full = (
            f"enum {self.get_typescript_interface_name(model)} {interface_body}"
        )

        if self.export_interface:
            interface_full = f"export {interface_full}"

        return interface_full

    def get_typescript_interface_name(self, model: OpenAPIProperty):
        if not model.title:
            raise ValueError(
                f"Model must have a title to retrieve its typescript name: {model}"
            )

        replace_chars = {" ", "[", "]"}
        values = model.title

        for char in replace_chars:
            values = values.strip(char)
            values = values.replace(char, "_")

        return camelize(values)

    def validate_typescript_candidate(self, model: Type[BaseModel]):
        """
        JSON only supports some types, so we need to validate that the given model not only
        is valid in Python but will also be valid when serialized over the wire.

        """
        for typehint in yield_all_subtypes(model):
            origin = get_origin(typehint)
            args = get_args(typehint)
            if origin and origin in {dict, Dict}:
                # We only support keys that are strings
                if args and not issubclass(args[0], str):
                    raise ValueError(
                        f"Key must be a string for JSON dictionary serialization. Received `{typehint}`."
                    )



import asyncio
from time import monotonic_ns

import pytest
from pydantic import BaseModel

from mountaineer.__tests__.common import calculate_primes
from mountaineer.cropper import FunctionCropException, crop_function_for_return_keys


class ExampleResponse(BaseModel):
    a: int
    b: int


def test_simple_sync_function():
    """
    Make sure that we successfully crop the AST to only compute
    logic necessary for our output
    """

    def example_function():
        x = 5
        y = 10
        a = x + y + calculate_primes(10000)
        b = x + calculate_primes(1000000)
        return {"a": a, "b": b}

    optimized_func = crop_function_for_return_keys(example_function, ["a"], locals())

    # Our raw function execution should be over 1s
    # Our optimized function should be less than 0.05s
    start = monotonic_ns()
    output_value = example_function()
    assert ((monotonic_ns() - start) / 1e9) > 1
    assert output_value == {"a": 1244, "b": 78503}

    start = monotonic_ns()
    output_value = optimized_func()
    assert ((monotonic_ns() - start) / 1e9) < 0.05
    assert output_value == {"a": 1244}


def test_conditional_sync_function():
    """
    Test our ability to parse if/else statements
    """

    def example_function(x: int):
        if x > 50:
            a = calculate_primes(10)
            b = calculate_primes(1000000)
        else:
            a = calculate_primes(0)
            b = calculate_primes(1000000)
        return {"a": a, "b": b}

    optimized_func = crop_function_for_return_keys(example_function, ["a"], locals())

    start = monotonic_ns()
    output_value = optimized_func(100)
    assert ((monotonic_ns() - start) / 1e9) < 0.05
    assert output_value == {"a": 4}

    start = monotonic_ns()
    output_value = optimized_func(0)
    assert ((monotonic_ns() - start) / 1e9) < 0.05
    assert output_value == {"a": 0}


def test_pydantic_model():
    def example_function():
        x = 5
        y = 10
        a = x + y + calculate_primes(10000)
        b = x + calculate_primes(1000000)
        return ExampleResponse(a=a, b=b)

    optimized_func = crop_function_for_return_keys(example_function, ["a"], locals())

    start = monotonic_ns()
    output_value = optimized_func()
    assert ((monotonic_ns() - start) / 1e9) < 0.05
    assert output_value == {"a": 1244}


@pytest.mark.asyncio
async def test_async_function():
    async def example_fn_short():
        await asyncio.sleep(0.001)
        return 5

    async def example_fn_long():
        await asyncio.sleep(10)
        return 10

    async def example_function_async():
        x = 5
        y = 10
        a = x + y + await example_fn_short()
        b = x + await example_fn_long()
        return ExampleResponse(a=a, b=b)

    optimized_func_async = crop_function_for_return_keys(
        example_function_async, ["a"], locals()
    )

    start = monotonic_ns()
    output_value = await optimized_func_async()
    assert ((monotonic_ns() - start) / 1e9) < 0.05
    assert output_value == {"a": 20}


def test_invalid_return_type():
    def example_function():
        return 5

    with pytest.raises(FunctionCropException):
        crop_function_for_return_keys(example_function, ["a"], locals())



{% if create_stub_files %}
from uuid import UUID, uuid4

from mountaineer import sideeffect, ControllerBase, Metadata, RenderBase
from mountaineer.database import DatabaseDependencies

from fastapi import Request, Depends
from pydantic import BaseModel
from sqlmodel import select
from sqlalchemy.ext.asyncio import AsyncSession

from {{project_name}} import models


class HomeRender(RenderBase):
    items: list[models.DetailItem]


class HomeController(ControllerBase):
    url = "/"
    view_path = "/app/home/page.tsx"

    async def render(
        self,
        session: AsyncSession = Depends(DatabaseDependencies.get_db_session)
    ) -> HomeRender:
        items = await session.execute(select(models.DetailItem))
        return HomeRender(
            items=items.scalars().all(),
            metadata=Metadata(title="Home"),
        )

    @sideeffect
    async def new_detail(
        self,
        session: AsyncSession = Depends(DatabaseDependencies.get_db_session)
    ) -> None:
        obj = models.DetailItem(description="Untitled Item")
        session.add(obj)
        await session.commit()
{% endif %}


{% if create_stub_files %}
from mountaineer.database import SQLModel, Field

class DetailItem(SQLModel, table=True):
    id: int | None = Field(default=None, primary_key=True)
    description: str
{% endif %}


"""
Utilities to generate some common Typescript objects from
Python definitions.

"""
from typing import Any

from mountaineer.client_builder.openapi import (
    OpenAPISchemaType,
    URLParameterDefinition,
    get_types_from_parameters,
)


class TSLiteral(str):
    """
    This string should not be quoted when used in a Typescript insert.
    """

    def __add__(self, other):
        # Concatenate self (TSLiteral) with other (any string)
        # and return a new TSLiteral object
        if isinstance(other, str):
            return TSLiteral(super(TSLiteral, self).__add__(other))
        return NotImplemented

    def __radd__(self, other):
        # Handle right-side addition, where a string is added to a TSLiteral
        if isinstance(other, str):
            return TSLiteral(other.__add__(self))
        return NotImplemented


def python_payload_to_typescript(payload: Any) -> str:
    """
    Take an element with python tokens that should be outputted to
    Typescript and consolidate them into a valid payload.

    """
    if isinstance(payload, dict):
        # Convert the children
        children_lines: list[str] = []
        for key, value in payload.items():
            if isinstance(value, TSLiteral):
                # If the literal is the same as the key, we don't need to include it since it's referencing
                # the same variable.
                if key == value:
                    children_lines.append(key)
                    continue

            key = python_payload_to_typescript(key)
            value = python_payload_to_typescript(value)
            children_lines.append(f"{key}: {value}")

        children_str = ",\n".join(children_lines)
        return f"{{\n{children_str}\n}}"
    elif isinstance(payload, list):
        children_lines = [python_payload_to_typescript(child) for child in payload]
        children_str = ",\n".join(children_lines)
        return f"[\n{children_str}\n]"
    elif isinstance(payload, TSLiteral):
        return payload
    elif isinstance(payload, str):
        return f"'{payload}'"
    elif isinstance(payload, bool):
        return str(payload).lower()
    elif isinstance(payload, (int, float)):
        return str(payload)
    elif payload is None:
        return "null"
    else:
        raise ValueError(
            f"Unknown payload type {type(payload) } for Typescript conversion."
        )


def map_openapi_type_to_ts(openapi_type: OpenAPISchemaType):
    mapping = {
        "string": "string",
        "integer": "number",
        "number": "number",
        "boolean": "boolean",
        "null": "null",
        "array": "Array<{types}>",
        "object": "{types}",
    }
    return mapping[openapi_type]


def get_typehint_for_parameter(parameter: URLParameterDefinition):
    """
    Get the typehint for a parameter, which may be a single type or a union of types.

    """
    parameter_types = set(get_types_from_parameters(parameter.schema_ref))

    key = TSLiteral(parameter.name) + (
        TSLiteral("?") if not parameter.required else TSLiteral("")
    )
    value = TSLiteral(
        " | ".join(
            # Sort helps with consistency of generated code
            sorted([map_openapi_type_to_ts(raw_type) for raw_type in parameter_types])
        )
    )

    return key, value



from pathlib import Path

from click import secho

from create_mountaineer_app.environments.base import EnvironmentBase
from create_mountaineer_app.environments.poetry import PoetryEnvironment
from create_mountaineer_app.environments.venv import VEnvEnvironment
from create_mountaineer_app.external import (
    has_npm,
    npm_install,
)
from create_mountaineer_app.generation import ProjectMetadata, format_template
from create_mountaineer_app.templates import get_template_path

IGNORE_FILES = {"__pycache__", "node_modules"}
ALLOW_HIDDEN_FILES = {
    # A template .env file is explicitly included in our build logic
    ".env",
    ".gitignore",
    ".vimrc",
    ".vscode",
}


def environment_from_metadata(metadata: ProjectMetadata) -> EnvironmentBase:
    if metadata.use_poetry:
        return PoetryEnvironment()
    else:
        return VEnvEnvironment()


def should_copy_path(root_path: Path, path: Path):
    """
    Determine whether we should copy the template path to our final project.
    We need to ignore certain build-time directories that don't actually
    have code logic.

    - Ignore explicitly ignored files
    - Ignore hidden files and folders

    """
    relative_path = path.relative_to(root_path)

    for part in relative_path.parts:
        if part in IGNORE_FILES:
            return False
        if part.startswith(".") and part not in ALLOW_HIDDEN_FILES:
            return False
    return True


def copy_source_to_project(template_base: Path, metadata: ProjectMetadata):
    template_paths = list(template_base.glob("**/*"))

    if not template_paths:
        all_template_paths = list(get_template_path("").glob("**/*"))
        secho(
            f"No templates found in {template_base}.\n"
            f"Local found: {template_paths}\n"
            f"All found: {all_template_paths}\n"
            "This might indicate an issue with your install or the pypi packaging pipeline.",
            fg="red",
        )
        raise Exception("No templates found.")

    for template_path in template_paths:
        if template_path.is_dir() or not should_copy_path(template_base, template_path):
            continue

        try:
            # Internally, format_template will re-look up the template
            output_bundle = format_template(template_path, template_base, metadata)
        except Exception as e:
            secho(f"Error formatting {template_path}: {e}", fg="red")
            raise e

        if not output_bundle.content.strip():
            secho(
                f"No content detected in {output_bundle.path}, skipping...", fg="yellow"
            )
            continue

        secho(f"Creating {output_bundle.path}")
        full_output = metadata.project_path / output_bundle.path
        full_output.parent.mkdir(parents=True, exist_ok=True)
        full_output.write_text(output_bundle.content)


def build_project(metadata: ProjectMetadata, install_deps: bool = True):
    template_base = get_template_path("project")

    copy_source_to_project(template_base, metadata)
    secho(f"Project created at {metadata.project_path}", fg="green")

    if install_deps:
        environment = environment_from_metadata(metadata)

        try:
            environment.install_project(metadata.project_path)
        except Exception as e:
            secho(f"Error installing python dependencies: {e}", fg="red")

        if has_npm():
            npm_install(metadata.project_path / metadata.project_name / "views")
        else:
            secho(
                "npm is not installed and is required to install React dependencies.",
                fg="red",
            )

        # Update the metadata now that we have a valid environment
        env_path = Path(environment.get_env_path(metadata.project_path))
        metadata.venv_base = str(env_path.parent)
        metadata.venv_name = env_path.name

        secho("Environment created successfully", fg="green")

    # Now copy the editor-specific files
    if metadata.editor_config:
        editor_template_base = (
            get_template_path("editor_configs") / metadata.editor_config.value
        )
        copy_source_to_project(editor_template_base, metadata)
        secho(f"Editor config created at {metadata.project_path}", fg="green")



from pathlib import Path
from unittest.mock import ANY
from uuid import uuid4

import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient
from pydantic import BaseModel

from mountaineer.app import AppController
from mountaineer.cli import (
    IsolatedBuildConfig,
    IsolatedEnvProcess,
    find_packages_with_prefix,
)
from mountaineer.controller import ControllerBase
from mountaineer.controllers.exception_controller import ExceptionController


@pytest.fixture
def env_process(tmpdir: str):
    """
    Custom isolated environment process with a valid exception controller
    that works without a build.

    """
    fake_ssr_template_path = Path(tmpdir) / f"{uuid4()}.html"
    fake_ssr_template_path.write_text(
        """
        const SSR = {
            x: () => `Passthrough: ${JSON.stringify(SERVER_DATA)}`
        }
        """
    )

    # We don't want to test our default exception logic, which requires a full build
    # to derive the ssr page
    exception_controller = ExceptionController()
    exception_controller.ssr_path = fake_ssr_template_path

    # Install our exception handler hook. The rest of the env process is unused, we just
    # need it for the mounted controller.
    # This is a hack to avoid calling run(), which will trigger an actual build
    env_process = IsolatedEnvProcess(IsolatedBuildConfig(webcontroller=""))
    env_process.exception_controller = exception_controller

    return env_process


@pytest.fixture
def invalid_build_app_controller(env_process: IsolatedEnvProcess, tmp_path: Path):
    (tmp_path / "page.tsx").write_text(
        """
        import React from "react";
        export default function Page() {
            return <div id="root">Value</div>;
        }
        """
    )
    (tmp_path / "package.json").write_text("{}")
    (tmp_path / "node_modules").mkdir()

    class ValidControllerInvalidBuild(ControllerBase):
        url = "/"
        view_path = "/page.tsx"

        def render(self) -> None:
            pass

    app_controller = AppController(view_root=tmp_path)
    app_controller.register(ValidControllerInvalidBuild())
    app_controller.app.exception_handler(Exception)(env_process.handle_dev_exception)

    return app_controller


def test_dev_thrown_exception_on_get(env_process: IsolatedEnvProcess):
    """
    Exceptions encountered during GET requests will affect the user's ability to render
    the page, so we should nicely format these on to the development server in-lieu of showing
    an empty page. Backend errors on the other hand are often used for validation and caught
    by the frontend, so we should not intercept these.

    """
    app = FastAPI()
    app.exception_handler(Exception)(env_process.handle_dev_exception)

    class EchoModel(BaseModel):
        value: str

    @app.get("/")
    def get():
        raise ValueError("This is a test")

    @app.post("/api", response_model=EchoModel)
    def post(content: EchoModel):
        return content

    with TestClient(app=app, raise_server_exceptions=False) as test_client:
        response = test_client.get("/")
        assert response.status_code == 500
        assert (
            '<div id="root">Passthrough: {"exception":"This is a test",'
            in response.text
        )

        response = test_client.post("/api", json={"value": 123})
        assert response.status_code == 422
        assert response.json() == {
            "detail": [
                {
                    "input": 123,
                    "loc": ["body", "value"],
                    "msg": ANY,
                    "type": "string_type",
                    "url": ANY,
                }
            ]
        }


def test_build_does_not_raise(
    invalid_build_app_controller: AppController,
    env_process: IsolatedEnvProcess,
):
    """
    Build errors in the isolated process shouldn't actually raise. If they do they
    will short-circuit the rest of our isolated watcher process (ie. won't actually
    bring up the `runserver`, which is not what we want).

    """
    env_process.run_build(invalid_build_app_controller)

    # Ensure that we've set the expected instance variable
    assert invalid_build_app_controller.build_exception is not None


def test_build_exception_on_get(
    invalid_build_app_controller: AppController,
    env_process: IsolatedEnvProcess,
):
    """
    Display build-time exceptions on the development server as well.

    """
    env_process.run_build(invalid_build_app_controller)

    # Also ensure that this exception is displayed on the development server
    with TestClient(
        app=invalid_build_app_controller.app, raise_server_exceptions=False
    ) as test_client:
        response = test_client.get("/")
        assert response.status_code == 500
        assert (
            # Thrown because we did not install any node modules in this test package
            'Could not resolve "react"' in response.text
        )


def test_find_packages_with_prefix():
    # Choose some packages that we know will be in the test environment
    assert set(find_packages_with_prefix("fasta")) == {"fastapi"}
    assert set(find_packages_with_prefix("pydan")) == {
        "pydantic",
        "pydantic_core",
        "pydantic-settings",
    }



from typing import Generic, TypeVar
from uuid import UUID, uuid4

import pytest
from pydantic import ValidationError

from mountaineer.database.sqlmodel import Field, SQLModel

T = TypeVar("T")


def test_generic_data_models():
    class GenericModel(SQLModel, Generic[T]):
        id: UUID = Field(default_factory=uuid4, primary_key=True)
        value: T
        default_value: str = "default"

    class MySubclass(GenericModel[int], table=True):
        pass

    obj = MySubclass(value=1)
    assert obj.value == 1

    with pytest.raises(ValidationError):
        obj = MySubclass(value="abc")  # type: ignore



import subprocess
from os import environ, pathsep
from pathlib import Path
from tempfile import NamedTemporaryFile

from click import secho

from create_mountaineer_app.environments.base import EnvironmentBase


class PoetryEnvironment(EnvironmentBase):
    """
    Supports a Poetry environment backend for our webapp.

    """

    def __init__(self):
        super().__init__()

        # While running our script within poetry, poetry will inject a POETRY_ACTIVE
        # environment variable. This will cause poetry to only install within the
        # current directory. We want to avoid this since the path we're given could be
        # anywhere in the system. We'll limit the scope of the new poetry subprocess to
        # only the PATH environment variable.
        self.limited_scope_env = {
            "PATH": environ["PATH"],
            **self.global_env,
        }

    def has_provider(self):
        try:
            # Attempt to get the version of poetry to check if it's installed
            subprocess.run(
                ["poetry", "--version"],
                check=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            return True
        except subprocess.CalledProcessError:
            # Poetry is installed but there might be a problem with it
            return True
        except FileNotFoundError:
            # Poetry is not installed
            return False

    def install_project(self, project_path: Path):
        subprocess.run(
            ["poetry", "install"],
            check=True,
            cwd=str(project_path),
            env=self.limited_scope_env,
        )

        # Retrieve the location of the new virtualenv
        env_path = self.get_env_path(project_path)
        secho(f"Poetry venv created: {env_path}", fg="green")

    def install_provider(self):
        """
        Attempt to install poetry via the official installation script:
        > curl -sSL https://install.python-poetry.org | python3 -

        Raises an exception if the installation fails.

        """
        with NamedTemporaryFile(delete=False) as tmp_file:
            subprocess.run(
                [
                    "curl",
                    "-sSL",
                    "https://install.python-poetry.org",
                    "-o",
                    tmp_file.name,
                ],
                check=True,
            )
            subprocess.run(["python3", tmp_file.name], check=True)

        secho("Poetry installed successfully.", fg="green")

    def add_poetry_to_path(self) -> None:
        """
        Mount poetry in the current terminal session and the user's shell profile.

        """
        directory = str(Path.home() / ".local" / "bin")

        shell = environ.get("SHELL")
        if not shell:
            secho("Could not detect the shell environment.", fg="red")
            raise Exception("Shell environment detection failed.")

        profile_path: Path
        if "bash" in shell:
            profile_path = Path("~/.bashrc")
        elif "zsh" in shell:
            profile_path = Path("~/.zshrc")
        else:
            secho(f"Unsupported shell: {shell}", fg="red")
            raise Exception("Unsupported shell.")

        # Expand the user's home directory and prepare the command to modify the profile file.
        profile_abs_path = profile_path.expanduser()
        path_command = f'export PATH="$PATH:{directory}"\n'

        # Check if the path is already in the file to avoid duplication
        with open(profile_abs_path, "r+") as profile:
            if path_command not in profile.read():
                profile.write(path_command)
                secho(f"Added {directory} to {profile_path}", fg="green")
            else:
                secho(f"{directory} is already in {profile_path}", fg="yellow")

        # Apply the path change to the current session
        environ["PATH"] += pathsep + directory
        secho("The PATH has been updated for the current session.", fg="green")

    def run_command(self, command: list[str], path: Path):
        return subprocess.Popen(
            ["poetry", "run", *command],
            cwd=path,
            env=self.limited_scope_env,
        )

    def get_env_path(self, project_path: Path) -> str:
        result = subprocess.run(
            ["poetry", "env", "info", "--path"],
            check=True,
            capture_output=True,
            cwd=str(project_path),
            env=self.limited_scope_env,
        )

        return result.stdout.decode().strip()



import asyncio
import socket
from functools import wraps
from typing import Any, Callable, Coroutine, TypeVar


async def gather_with_concurrency(
    tasks: list[Coroutine[Any, Any, Any]],
    *,
    n: int,
    catch_exceptions: bool = False,
) -> list[Any]:
    semaphore = asyncio.Semaphore(n)

    async def sem_task(task: Coroutine[Any, Any, Any]) -> Any:
        async with semaphore:
            return await task

    return await asyncio.gather(
        *(sem_task(task) for task in tasks),
        return_exceptions=catch_exceptions,
    )


T = TypeVar("T")


def async_to_sync(async_fn: Callable[..., Coroutine[Any, Any, T]]) -> Callable[..., T]:
    @wraps(async_fn)
    def wrapper(*args, **kwargs) -> T:
        loop = asyncio.get_event_loop()
        result = loop.run_until_complete(async_fn(*args, **kwargs))
        return result

    return wrapper


def get_free_port() -> int:
    """
    Leverage the OS-port shortcut :0 to get a free port. Return the value
    of the port that was assigned.

    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("", 0))
        port = s.getsockname()[1]
        s.close()
    return int(port)
